// Name: main.c
// char    MYD_id[]=
// " sphDESC.C Ver.051 2010/09/23  MORIYOSHI DENSHI CO.,LTD. COPYRIGHT(C) MAR.2009 ";
// Ver.034 + USI TWI communication --> Ver.039 (2010B) + first CRC --> Ver.051 (+8B)
// License: GNU GPL v2
// avr Tiny2313 12MHz

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <string.h>

#include "hidtermconfig.h"

#include "usbdrv.h"
#if (USB_CFG_DESCR_PROPS_UNKNOWN & USB_PROP_IS_DYNAMIC)
static  uchar serbuf[32+0]; // =0;
#define SerRoutine 1
#endif

#include "usbdrv.c"

//static uchar reportId;
#if SerRoutine
#define offsetR UBRRL
//@ (-8)
#else
static uchar offset; // data already processed in current transfer
#endif

static uchar usbData[33];
//static uchar page_mode;
#define page_mode GPIOR0

typedef struct shortaddr{
    uchar   lbyte;
    uchar   hbyte;
}shrtaddr_t;

union{
    uchar       bytes[1];
    uint16_t    w;
    shrtaddr_t  b;
}               page_addr;

PROGMEM char usbHidReportDescriptor[33] = {
    0x06, 0x00, 0xff,              // USAGE_PAGE (Generic Desktop)
    0x09, 0x01,                    // USAGE (Vendor Usage 1)
    0xa1, 0x01,                    // COLLECTION (Application)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x26, 0xff, 0x00,              //   LOGICAL_MAXIMUM (255)
    0x75, 0x08,                    //   REPORT_SIZE (8)

    0x85, 0x02,                    //   REPORT_ID (2)
    0x95,    7,                    //   REPORT_COUNT (7)
    0x09, 0x00,                    //   USAGE (Undefined)
    0xb2, 0x02, 0x01,              //   FEATURE (Data,Var,Abs,Buf)

    0x85, 0x04,                    //   REPORT_ID (4)
    0x95,   39,                    //   REPORT_COUNT (39)
    0x09, 0x00,                    //   USAGE (Undefined)
    0xb2, 0x02, 0x01,              //   FEATURE (Data,Var,Abs,Buf)

    0xc0                           // END_COLLECTION
};


#define wait GPIOR1
//static uchar wait; // 60 160
static void delay(uchar d) {
    while(d--);
}

static uint8_t usi_trans(uchar data) {
    USICR = (1 << USIWM0) | (1 << USICS1) | (1 << USICLK); //MYD100923
    USIDR = data;
    //USISR |= (1<<USIOIF); //MYD090925
    USISR = (1<<USISIF) | (1<<USIOIF) | (1<<USIPF) | 0; //MYD100923
    do{
        delay(wait);
        USICR = (1 << USIWM0) | (1 << USICS1) | (1 << USICLK) | (1 << USITC);
    } while(!(USISR & (1 << USIOIF)) );
    return USIDR;
}

uchar   usbFunctionSetup(uchar data[8]) {
usbRequest_t    *rq = (void *)data;
    uchar rval; //MYD091021

    rval=0;
    if(rq->bRequest == USBRQ_HID_SET_REPORT){
//      reportId = rq->wValue.bytes[0]; // store report ID
#if SerRoutine
        offsetR = 0; //@
#else
        offset = 0;
#endif
        rval = 0xFF;
    }else if(rq->bRequest == USBRQ_HID_GET_REPORT){
        usbMsgPtr = usbData;
        usbData[0] = rq->wValue.bytes[0]; // report ID
        rval = 8;
        if(usbData[0] == 4) {
            rval = 40;
        }
    }
    return rval; //MYD091021 (-6)
}

uchar usbFunctionWrite(uchar *data, uchar len) {
    uchar i, d1, d2;
#if SerRoutine
    uchar offset;

    offset = offsetR; //@
    //@ (-6) static uchar tcnt;
#define tcnt UBRRH
#else
    static uchar tcnt; //MYD091021 (-6)
#endif

    d2 = data[2]; //@ (-0)
    d1 = data[1]; //@ (-0)

    if(offset == 0 && !(d1 & 0x40)) {
//  if(data[1] == 1) // TEST
        usbData[1] = d2;
        if(d1 == 2) { // SET_LED
//          PORTD = (PORTD & ~3) | (data[2] & 3);
//          PORTB = (PORTB & ~0x10) | (data[3] & 0x10); // ~reset
//          PORTB = (PORTB & ~0x0C) | ((data[2] << 2) & 0x0C); // ~activeLED ~powerLED
//          PORTB = (PORTB & ~0x1C) | ((data[3] & 0x10) | ((data[2] << 2) & 0x0C));
#if SerRoutine
            USICR = (1<<USIWM1) | (1<<USIWM0) | (1<<USICS1);
#else
            USICR = 0;
#endif
            PORTB = data[3]; //@ (-12) | ((data[4] & (1<<PB7)) ? 0 : 0xE0);
            //if(data[4] == 0x55) DDRB = data[5]; // (+1)
            if(data[4] & (1<<PB2)) //@ (-4)
                     DDRB = data[4];
#if !SerRoutine
                else DDRB |= (1<<PB7); // (+6)
#endif
            //if(!(PORTB & (1<<PB4))) DDRB |= (1<<PB7); // ~reset Low (+2)
//          UCSRB = data[4]; // USART control, only Face
//          PORTD = data[5]; // only Face
#if 1
            usbData[2] = PIND; //@ (+4)
#endif
        }else if(d1 == 18) { // Set wait
            wait = d2;
        }else if(d1 == 16) { // SPI
            for(i = 0; i < 4; i++) {
                usbData[i+1] = usi_trans(data[i+2]);
            }
        }else if(d1 == 20) { // Page set
            page_mode = d2;
            page_addr.b.hbyte = data[3];
            page_addr.b.lbyte = data[4];
        }
    }else{
        if(offset == 0) { // Page buf
            tcnt = (d1 & 0x07) + 2;
            i = 2;
        }else /* if(!(reportId == 4 && offset < 32)) tcnt = 0; else */ {
            tcnt = (offset > 32-8) ? 32-offset : 8;
            i = 0;
        }
        while(i < tcnt) {
            usi_trans(page_mode);
            usi_trans(page_addr.b.hbyte);
            usi_trans(page_addr.b.lbyte);
            usbData[++offset] = usi_trans(data[i++]);
            if(page_mode & 0x88) { // EEPROM or Flash(P=Hi)
                page_addr.w++;
                page_mode &= ~0x08;
            }else{
                page_mode |= 0x08;
            }
        }
    }
#if SerRoutine
    offsetR = offset; //@
#endif
    return 1;
}

#if SerRoutine
#define UsbRetSts serbuf[0]
#define SerRXbuff serbuf[1]
#define UsbwV0    serbuf[16]
#define UsbwV1    serbuf[17]
#define UsbwI0    serbuf[18]
#define UsbwL1    serbuf[21]
#define SerTXbuff serbuf[22]

// use at host(Hub driver) IOCTL_USB_GET_DESCRIPTOR_FROM_NODE_CONNECTION
/*uchar usbFunctionDescriptor(struct usbRequest *rq) { // bRequest==6 GET_DESCRIPTOR
    uchar i, *pser, *prq; pser = &serbuf[16]; prq = &rq->wValue.bytes[0]; i=6;
    do *pser++ = *prq++; while(--i);
//  serbuf[6+16] = rq->bmRequestType;
//  serbuf[0+16] = rq->wValue.bytes[0];
//  serbuf[1+16] = rq->wValue.bytes[1];
//  serbuf[2+16] = rq->wIndex.bytes[0];
//  serbuf[3+16] = rq->wIndex.bytes[1];
//  serbuf[4+16] = rq->wLength.bytes[0];
//  serbuf[5+16] = rq->wLength.bytes[1];
    usbMsgPtr = serbuf; // &serbuf[16]
    return 7;
}*/

#endif

int main(void) {
#if SerRoutine
    uchar usstxd, ussrxd;
    //uchar usslvsts=0; // b7:exist send data b6:exist receive data b5:command received (-10)
#define usslvsts GPIOR2
#endif

    // initialize hardware
    PORTD = (1<<PD6) | (1<<PD5) | (1<<PD1) | (1<<PD0);  // pullups
                       // no pullups on USB pins
    DDRD = 0x7F & ~USBMASK & ~(1<<PD6) & ~(1<<PD5) & ~(1<<PD2) & ~(1<<PD1) & ~(1<<PD0);
                       // all inputs except USB data
#if SerRoutine
    USICR = (1<<USIWM1) | (1<<USIWM0) | (1<<USICS1);
    /*PORTB |= (1<<PB5); // for two wire slave USI
    PORTB |= (1<<PB7);
    DDRB |= (1<<PB7);*/
    PORTB = 0xF8;      // no pullup PB7,5 on USI // ~RST Hi, activeLED off
    DDRB = (1<<PB7) | (1<<PB6) | (1<<PB4) | (1<<PB3) | (1<<PB2);
    USISR = (1<<USISIF) | (1<<USIOIF) | (1<<USIPF) | 1;
#else
    PORTB = 0xF8;      // pullups PB7,6,5,1,0
    DDRB = (1<<PB4) | (1<<PB3) | (1<<PB2);
//  DDRB = ~(1<<PB5);  // all outputs except USI input data for Ver.010
#endif

    usbInit();
    sei();
    while(1) { // main event loop
        usbPoll();

#if SerRoutine
//void serLoop()
{   uchar retStsRX, retStsTX; uchar *dest, *src;
    retStsRX = UsbRetSts & 0x0F;
    retStsTX = UsbRetSts >> 4;

    if(UsbwV1 & 0x80) { uchar i;
        UsbwV1 &= 0x7F;
        if(UsbwV1 == 0x60) { // PORTB set & TXbuffer flush
            retStsTX = 0; //@ (+2)
            PORTB = UsbwV0; //@ (+8)
            usslvsts &= ~(1<<7); //@ (+2)
#if 1
            if(retStsRX < 6) serbuf[6] = PIND; //@ (+10)
#endif
        }else{
//@         if( /*(i = (UsbwV1 & 0x07)) &&*/ (retStsTX < 7) ) {
            if( (i = (UsbwV1 & 0x07)) && (retStsTX < 7) ) { //@
                UsbwL1 = UsbwV0;
                dest = &SerTXbuff + retStsTX; src = &UsbwI0;
//@             i = UsbwV1 & 0x07;
                retStsTX += i;
                do *dest++ = *src++; while(--i); //@ i--
            }
            if(retStsRX >= 7) {
                retStsRX -= 6;
                dest = &SerRXbuff; src = &SerRXbuff+6;
lc:             i = 9;
                do *dest++ = *src++; while(--i);
            }else{
                retStsRX = 0;
            }
        }
    }
    if( (usslvsts & (1<<6)) && (retStsRX < 15) ) {
        *(&SerRXbuff + (retStsRX++)) = ussrxd;
        usslvsts &= ~(1<<6);
    }
    if(retStsTX && !(usslvsts & (1<<7)) ) {
        usstxd = SerTXbuff;
        usslvsts |= (1<<7);
        --retStsTX;
        dest = &SerTXbuff; src = &SerTXbuff+1;
        goto lc;
//      i = 9;
//      do *dest++ = *src++; while(--i);
    }

    UsbRetSts = ((retStsTX << 4) & 0xFF) | retStsRX; //@@ (-10)
}
//void  ussint()
{
    if(USISR & (1<<USIOIF)) { // counter over flow
        DDRB &= ~(1<<PB5);
        //USISR = 0;
        USISR |= (1<<USIOIF);
    }
    if(USISR & (1<<USISIF)) { // start condition
        if(USISR & (1<<USIPF)) { // stop condition :: data received
            ussrxd = USIDR;
            usslvsts |= (1<<6);
        }else{
            if(PORTB & (1<<4)) DDRB |= (1<<PB5);
            if(USIDR & (1<<1)) {
#if 1
                uchar tmp;
                tmp = (PIND & 0x23); // (+10)
                if(PIND & 0x40) tmp |= (1<<2); // (+4)
                tmp |= usslvsts;
                tmp ^= 0xC0;
                USIDR = tmp;
                //USIDR = ( (tmp=((PIND & 0x23) | ((PIND & 0x40) ? 4 : 0))) | usslvsts ) ^ 0xC0;
#else
                USIDR = usslvsts ^ 0xC0; // (-10)
#endif
            }else{
                USIDR = usstxd;
                usslvsts &= ~(1<<7);
            }
        }
        USISR = (1<<USIOIF) | (1<<USIPF) | 0;
        USISR |= (1<<USISIF);
    }
}
#endif

    }
}
