// Name: main.c
// char    MYD_id[]=
// " sphLtest.C Ver.050 2010/09/18  MORIYOSHI DENSHI CO.,LTD. COPYRIGHT(C) OCT.2009 ";
// License: GNU GPL v2
// avr Tiny2313 12MHz
// fuse L:E0(11100000) H:DF(11011111) E:01(00000001) // External Clock
// fuse L:FF(11111111) H:DF(11011111) E:01(00000001) // XTAL
//   or L:CF(11001111) H:DF(11011111) E:01(00000001) // CEROLOCK

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <string.h>

#ifndef uchar
#define uchar   unsigned char
#endif

#define WORD unsigned int

#define tty 0

uchar usstxd, ussrxd;//, usscmnd;
uchar usslvsts=0; // b7:exist send data b6:exist receive data

void  coninit() {
    // initialize hardware
#if tty
    PORTA = 0x02;    // PA1 pullup
    //DDRA = (1<<PA1); // PA1 output
    PORTD = 0x7F;    // all pullups include PD0(RXD), except PD1,PD3 high output
    DDRD = (1<<PD1)  // all inputs include RXD(PD0 TxD),PD4(~RTS)
         | (1<<PD3); // except TXD(PD1 RxD),PD3(~CTS) output
    PORTB = 0x00;    // no pullup
    DDRB = 0x1F;     // all outputs except PB7,PB6,PB5

    // USART configuration : 9600bps 8N1
    UBRRL  = 78-1;
    UCSRB  = (1<<RXEN) | (1<<TXEN) | (1<<RXCIE);
#else
    PORTA = 0x02;    // PA1 pullup
    PORTD = 0x7F;    // all pullups
    PORTB = 0x00;    // no pullup
    DDRB = 0x1F;     // all outputs except PB7,PB6,PB5
#endif
}

void  ussinit() { // USI TWI mode & SCL to auto-Low by detection of counter overflow
    USICR = (1<<USIWM1) | (1<<USIWM0) | (1<<USICS1);
    PORTB |= (1<<PB5); // set is no drived open-collector output
    PORTB |= (1<<PB7); // set is no drived open-collector output
    DDRB |= (1<<PB7);
    USISR = (1<<USISIF) | (1<<USIOIF) | (1<<USIPF) | 1;
}

void  cmndprcs(uchar cmnd) { uchar tmp;
    cmnd >>= 3;
    cmnd &= 0x1F;
    tmp = PORTB;
    tmp &= 0xE0;
    tmp |= cmnd;
    PORTB = tmp;
}

void  ussint1() {
    if(USISR & (1<<USIOIF)) { // counter over flow
        DDRB &= ~(1<<PB5);
        //USISR = 0;
        USISR |= (1<<USIOIF);
    }
    if(USISR & (1<<USISIF)) { // start condition
        if(USISR & (1<<USIPF)) { // stop condition :: data received
            ussrxd = USIDR;
            usslvsts |= (1<<6);
        }else{
            if(USIDR & (1<<1)) { uchar tmp;
                DDRB |= (1<<PB5);
                tmp = PIND;
                tmp >>= 1;
                tmp &= 0x3E;
                if(PINA & (1<<1)) tmp |= (1<<0);
                tmp |= usslvsts;
                tmp ^= 0xC0;
                USIDR = tmp;
            }else{
                if(USIDR & (1<<2)) {
                    cmndprcs(USIDR);
                }else{
                    DDRB |= (1<<PB5);
                    USIDR = usstxd;
                    usslvsts &= ~(1<<7);
                }
            }
        }
        USISR = (1<<USIOIF) | (1<<USIPF) | 0;
        USISR |= (1<<USISIF);
    }
}

uchar conins() { // return val. : ready
    ussint1();
    return (usslvsts & (1<<6));
}

uchar coninp() {
    while( !conins()) ;
    usslvsts &= ~(1<<6);
    return ussrxd;
}

uchar conots() { // return val. : busy
    ussint1();
    return (usslvsts & (1<<7));
}

void  conout(uchar dat) {
    while( conots()) ;
    usstxd = dat;
    usslvsts |= (1<<7);
}

#if tty
uchar ttysts, ttyibf;

/*void  ttyintr() {
    if(ttysts) {
        if(!(PIND & (1<<PD3)))
            PORTD |= (1<<PD3);
    }else{
        if(UCSRA & (1<<RXC)) {
            ttyibf = UDR;
            ttysts = 1;
            PORTD |= (1<<PD3);
        }else{
            PORTD &= ~(1<<PD3);
        }
    }
}*/

uchar ttyins() { // return val. : ready
//  ttyintr();
    cli();
    if(!ttysts && !(UCSRA & (1<<RXC)) ) PORTD &= ~(1<<PD3);
    sei();
    return ttysts;
}

uchar ttyinp() { uchar dat;
    while( !ttyins()) ;
    cli();
    ttysts = 0;
    dat = ttyibf;
    PORTD &= ~(1<<PD3);
    UCSRB |= (1<<RXCIE);
    sei();
//  ttyintr();
    return dat;
}

uchar ttyots() { // return val. : not ready
//  ttyintr();
    return (~UCSRA & (1<<UDRE)) | (PIND & (1<<PD4));
}

void  ttyout(uchar dat) {
    while( ttyots()) ;
    UDR = dat;
}
#endif

void  crlf() {
//  conout('\r'); // 0x0D enter
    conout('\n'); // 0x0A ^J
}

void  hxd(uchar dat) {
    dat &= 0x0F;
    if(dat >= 10) dat += ('A'-'0'-10);
    conout(dat + '0');
}

void  lbyte(uchar dat) {
    hxd(dat >> 4);
    hxd(dat);
}

void  ladr(WORD dat) {
    lbyte(dat >> 8);
    lbyte(dat & 0x00FF);
}

void  mesot(char *str) { uchar dat;
    while(1) {
        dat = pgm_read_byte(str);
        if(!dat) break;
        conout(dat); str++;
    }
}

uchar txtin() { uchar dat;
    dat = coninp();
    if(dat < 0x20) return dat; // control code
    conout(dat);
    if(dat >= 'a' && dat < 'z'+1) dat -= 'a'-'A';
    return dat;
}

uchar nibbl(uchar dat) {
    dat -= '0';
    if(dat < 10) return dat;
    dat += '0'-'A'+10;
    if(dat >= 10 && dat < 16) return dat;
    return 0x10;
}

extern uchar setdlmnbl(uchar *delmt, uchar ch);
/*uchar setdlmnbl(uchar *delmt, uchar ch) {
    *delmt = ch;
    return nibbl(ch);
}*/

WORD  expr1(uchar *delmt) { WORD dat; uchar ch;
    dat = 0;
    while(1) {
        ch = txtin();
//      *delmt = ch;
//      ch = nibbl(ch);
        ch = setdlmnbl(delmt, ch);
        if(ch >= 0x10) break;
        dat <<= 4;
        dat |= ch;
    }
    return dat;
}

PROGMEM char headmes[] = "avr HS-monitor\n";

int main(void) {
    uchar ch;

    coninit();
    ussinit();
    sei();

ts: while(1) {
        ch = coninp();
        if(ch == 0x1B) break; // esc
        conout(ch);
        if(ch != '\n') conout(ch); // 0x0A
    }
    crlf();
    mesot((char *)headmes);
    while(1) {
        conout('.');
        ch = txtin();
        if(ch == 0x03) goto ts; // ^C
        if(ch == 'I') { WORD adr;
            adr = expr1(&ch);
            if(ch < 0x20) conout(' ');
            ladr(adr);
            conout('=');
            lbyte(_MMIO_BYTE(adr)); // 0x20-5F:IO 0x60-DF:RAM
            if(ch == '\r') crlf();
        }else{
            if(ch == '\r') crlf();
            if(ch == '\n') conout(' ');
            conout('x');
            lbyte(ch);
        }
    }
}
